import json
import logging
from urllib.parse import quote

import requests

from pypuppetdb.errors import APIError, EmptyResponseError

log = logging.getLogger(__name__)

ENDPOINTS = {
    "facts": "pdb/query/v4/facts",
    "fact-names": "pdb/query/v4/fact-names",
    "nodes": "pdb/query/v4/nodes",
    "resources": "pdb/query/v4/resources",
    "catalogs": "pdb/query/v4/catalogs",
    "mbean": "metrics/v1/mbeans",
    # metrics v2 endpoint is now the jolokia library and all of its operations
    # https://jolokia.org/reference/html/protocol.html#jolokia-operations
    "metrics": "metrics/v2/read",
    "metrics-base": "metrics/v2",
    "metrics-exec": "metrics/v2/exec",
    "metrics-list": "metrics/v2/list",
    "metrics-search": "metrics/v2/search",
    "metrics-write": "metrics/v2/write",
    "metrics-version": "metrics/v2/version",
    "reports": "pdb/query/v4/reports",
    "events": "pdb/query/v4/events",
    "event-counts": "pdb/query/v4/event-counts",
    "aggregate-event-counts": "pdb/query/v4/aggregate-event-counts",
    "server-time": "pdb/meta/v1/server-time",
    "version": "pdb/meta/v1/version",
    "environments": "pdb/query/v4/environments",
    "factsets": "pdb/query/v4/factsets",
    "fact-paths": "pdb/query/v4/fact-paths",
    "fact-contents": "pdb/query/v4/fact-contents",
    "edges": "pdb/query/v4/edges",
    "pql": "pdb/query/v4",
    "inventory": "pdb/query/v4/inventory",
    "status": "status/v1/services/puppetdb-status",
    "cmd": "pdb/cmd/v1",
}

PARAMETERS = {
    "order_by": "order_by",
    "include_total": "include_total",
    "count_by": "count_by",
    "counts_filter": "counts_filter",
    "summarize_by": "summarize_by",
    "server_time": "server_time",
}

COMMAND_VERSION = {
    "deactivate node": 3,
    "replace catalog": 9,
    "replace facts": 5,
    "store report": 8,
}

ERROR_STRINGS = {
    "timeout": "Connection to PuppetDB timed out on",
    "refused": "Could not reach PuppetDB on",
}


class BaseAPI:
    """This is a Base or Abstract class and is not meant to be instantiated
    or used directly.

    The BaseAPI object defines a set of methods that can be
    reused across different versions of the PuppetDB API. If querying for a
    certain resource is done in an identical fashion across different versions
    it will be implemented here and should be overridden in their respective
    versions if they deviate.

    If :attr:`ssl` is set to `True` but either :attr:`ssl_key` or\
            :attr:`ssl_cert` are `None` this will raise an error.

    :param host: (optional) Hostname or IP of PuppetDB.
    :type host: :obj:`string`

    :param port: (optional) Port on which to talk to PuppetDB.
    :type port: :obj:`int`

    :param ssl_verify: (optional) Verify PuppetDB server certificate.
    :type ssl_verify: :obj:`bool` or :obj:`string`

    :param ssl_key: (optional) Path to our client secret key.
    :type ssl_key: :obj:`None` or :obj:`string` representing a filesystem\
            path.

    :param ssl_cert: (optional) Path to our client certificate.
    :type ssl_cert: :obj:`None` or :obj:`string` representing a filesystem\
            path.

    :param timeout: (optional) Number of seconds to wait for a response.
    :type timeout: :obj:`int`

    :param protocol: (optional) Explicitly specify the protocol to be used
            (especially handy when using HTTPS with ssl_verify=False and
            without certs)
    :type protocol: :obj:`None` or :obj:`string`

    :param url_path: (optional) The URL path where PuppetDB is served
            (if not at the root / path)
    :type url_path: :obj:`None` or :obj:`string`

    :param username: (optional) The username to use for HTTP basic
            authentication
    :type username: :obj:`None` or :obj:`string`

    :param password: (optional) The password to use for HTTP basic
            authentication
    :type password: :obj:`None` or :obj:`string`

    :param token: (optional) The X-auth token to use for X-Authentication
    :type token: :obj:`None` or :obj:`string`

    :param metric_api_version: (Default 'v2') Version of the metric API we're initialising.
    :type metric_api_version: :obj:`None` or :obj:`string`

    :raises: :class:`~pypuppetdb.errors.ImproperlyConfiguredError`
    """

    def __init__(
        self,
        host="localhost",
        port=8080,
        ssl_verify=True,
        ssl_key=None,
        ssl_cert=None,
        timeout=10,
        protocol=None,
        url_path=None,
        username=None,
        password=None,
        token=None,
        metric_api_version=None,
    ):
        """Initialises our BaseAPI object passing the parameters needed in
        order to be able to create the connection strings, set up SSL and
        timeouts and so forth."""

        self.api_version = "v4"
        if metric_api_version is not None and metric_api_version not in ["v1", "v2"]:
            raise ValueError(
                "metric_api_version specified must be None, 'v1' or 'v2',"
                " was given: '{}'".format(metric_api_version)
            )
        self.metric_api_version = metric_api_version if metric_api_version else "v2"
        self.host = host
        self.port = port
        self.ssl_verify = ssl_verify
        self.ssl_key = ssl_key
        self.ssl_cert = ssl_cert
        self.timeout = timeout
        self.token = token

        # Standardise the URL path to a format similar to /puppetdb
        if url_path:
            if not url_path.startswith("/"):
                url_path = "/" + url_path
            if url_path.endswith("/"):
                url_path = url_path[:-1]
        else:
            url_path = ""

        self.url_path = url_path

        self.session = requests.Session()

        if username and password:
            self.session.auth = (username, password)

        self.session.headers = {
            "content-type": "application/json",
            "accept": "application/json",
            "accept-charset": "utf-8",
        }

        if self.token:
            self.session.headers["X-Authentication"] = self.token

        if protocol is not None:
            protocol = protocol.lower()
            if protocol not in ["http", "https"]:
                raise ValueError("Protocol specified must be http or https")
            self.protocol = protocol
        elif self.ssl_key is not None and self.ssl_cert is not None:
            self.protocol = "https"
        elif self.token is not None:
            self.protocol = "https"
        else:
            self.protocol = "http"

    def disconnect(self):
        """Close all connections that this class opened up."""
        # If we don't explicitly close connections, we might cause other
        # functions or libraries to hang on the open connections. This happens
        # for example with using paramiko to tunnel PuppetDB connections
        # through ssh.
        self.session.close()

    def __enter__(self):
        """Set up environment for 'with' statement."""
        # Once this class has been instantiated, there's nothing more required
        return self

    def __exit__(self, type, value, trace):
        """Tear down connections."""
        self.disconnect()

    @property
    def version(self):
        """The version of the API we're querying against.

        :returns: Current API version.
        :rtype: :obj:`string`"""
        return self.api_version

    @property
    def base_url(self):
        """A base_url that will be used to construct the final
        URL we're going to query against.

        :returns: A URL of the form: ``proto://host:port``.
        :rtype: :obj:`string`
        """
        return "{proto}://{host}:{port}{url_path}".format(
            proto=self.protocol,
            host=self.host,
            port=self.port,
            url_path=self.url_path,
        )

    @property
    def total(self):
        """The total-count of the last request to PuppetDB
        if enabled as parameter in _query method

        :returns Number of total results
        :rtype :obj:`int`
        """
        if self.last_total is not None:
            return int(self.last_total)

    @staticmethod
    def _normalize_resource_type(type_):
        """Normalizes the type passed to the api by capitalizing each part
        of the type. For example:

        sysctl::value -> Sysctl::Value
        user -> User

        """
        return "::".join([s.capitalize() for s in type_.split("::")])

    def _url(self, endpoint, path=None):
        """The complete URL we will end up querying. Depending on the
        endpoint we pass in  this will result in different URL's with
        different prefixes.

        :param endpoint: The PuppetDB API endpoint we want to query.
        :type endpoint: :obj:`string`
        :param path: An additional path if we don't wish to query the\
                bare endpoint.
        :type path: :obj:`string`

        :returns: A URL constructed from :func:`base_url` with the\
                apropraite API version/prefix and the rest of the path added\
                to it.
        :rtype: :obj:`string`
        """

        log.debug("_url called with endpoint: {} and path: {}".format(endpoint, path))

        try:
            endpoint = ENDPOINTS[endpoint]
        except KeyError:
            # If we reach this we're trying to query an endpoint that doesn't
            # exist. This shouldn't happen unless someone made a booboo.
            raise APIError

        url = "{base_url}/{endpoint}".format(
            base_url=self.base_url,
            endpoint=endpoint,
        )

        if path is not None:
            url = f"{url}/{quote(path)}"

        return url

    def _query(
        self,
        endpoint=None,
        path=None,
        query=None,
        order_by=None,
        limit=None,
        offset=None,
        include_total=False,
        summarize_by=None,
        count_by=None,
        count_filter=None,
        payload=None,
        request_method="GET",
    ):
        """This method prepares a non-PQL query to PuppetDB. Actual making
        the HTTP request is done by _make_request().

        :param endpoint: (optional) The PuppetDB API endpoint we want to query.
                        Unnecessary when using PQL.
        :type endpoint: :obj:`string`
        :param path: An additional path if we don't wish to query the\
                bare endpoint.
        :type path: :obj:`string`
        :param query: (optional) An AST query to further narrow down the resultset.
        :type query: :obj:`string`
        :param order_by: (optional) Set the order parameters for the resultset.
        :type order_by: :obj:`string`
        :param limit: (optional) Tell PuppetDB to limit it's response to this\
                number of objects.
        :type limit: :obj:`int`
        :param offset: (optional) Tell PuppetDB to start it's response from\
                the given offset. This is useful for implementing pagination\
                but is not supported just yet.
        :type offset: :obj:`string`
        :param include_total: (optional) Include the total number of results
        :type order_by: :obj:`bool`
        :param summarize_by: (optional) Specify what type of object you'd like\
                to see counts at the event-counts and aggregate-event-counts \
                endpoints
        :type summarize_by: :obj:`string`
        :param count_by: (optional) Specify what type of object is counted
        :type count_by: :obj:`string`
        :param count_filter: (optional) Specify a filter for the results
        :type count_filter: :obj:`string`
        :param payload: (optional) Arbitrary payload to send as part of the request.
        :type payload: :obj:`dict`

        :raises: :class:`~pypuppetdb.errors.EmptyResponseError`

        :returns: The decoded response from PuppetDB
        :rtype: :obj:`dict` or :obj:`list`
        """

        # inside the list comprehension the locals()'s value changes
        # so we need to make a copy of the function's local() to use it there
        function_locals = locals().copy()
        log.debug(
            "_query called with: "
            +
            # comma-separated list of method arguments with their values
            ", ".join(
                [
                    f"{arg}={function_locals.get(arg, 'None')}"
                    for arg in function_locals.keys()
                    if arg != "self"
                ]
            )
        )

        if not endpoint:
            log.error("Endpoint is required!")
            raise APIError

        if payload is None:
            payload = {}

        url = self._url(endpoint, path=path)
        if query is not None:
            payload["query"] = query
        if order_by is not None:
            payload[PARAMETERS["order_by"]] = order_by
        if limit is not None:
            payload["limit"] = limit
        if include_total is True:
            payload[PARAMETERS["include_total"]] = json.dumps(include_total)
        if offset is not None:
            payload["offset"] = offset
        if summarize_by is not None:
            payload[PARAMETERS["summarize_by"]] = summarize_by
        if count_by is not None:
            payload[PARAMETERS["count_by"]] = count_by
        if count_filter is not None:
            payload[PARAMETERS["counts_filter"]] = count_filter

        return self._make_request(url, request_method, payload)

    def _make_request(self, url, request_method, payload):
        """
        Makes a GET or POST HTTP request to PuppetDB. If PuppetDB can be
        reached and answers within the timeout we'll decode the response
        and give it back or raise for the HTTP Status Code PuppetDB gave back.

        :param url: Complete URL to call
        :param request_method: GET or POST
        :param payload: data to send as parameters (GET) or in the body (POST)
        :return: response body as JSON
                 or raises an EmptyResponseError exception if it's empty
        """

        if request_method.upper() not in ["GET", "POST"]:
            log.error(f"Only GET or POST supported, {request_method} unsupported")
            raise APIError

        try:
            if request_method.upper() == "GET":
                r = self.session.get(
                    url,
                    params=payload,
                    verify=self.ssl_verify,
                    cert=(self.ssl_cert, self.ssl_key),
                    timeout=self.timeout,
                )
            else:
                r = self.session.post(
                    url,
                    data=json.dumps(payload, default=str),
                    verify=self.ssl_verify,
                    cert=(self.ssl_cert, self.ssl_key),
                    timeout=self.timeout,
                )

            r.raise_for_status()

            # get total number of results if requested with include-total
            # just a quick hack - needs improvement
            if "X-Records" in r.headers:
                self.last_total = r.headers["X-Records"]
            else:
                self.last_total = None

            json_body = r.json()
            if json_body is not None:
                return json_body
            else:
                del json_body
                raise EmptyResponseError

        except requests.exceptions.Timeout:
            log.error(
                "{} {}:{} over {}.".format(
                    ERROR_STRINGS["timeout"],
                    self.host,
                    self.port,
                    self.protocol.upper(),
                )
            )
            raise

        except requests.exceptions.ConnectionError:
            log.error(
                "{} {}:{} over {}.".format(
                    ERROR_STRINGS["refused"],
                    self.host,
                    self.port,
                    self.protocol.upper(),
                )
            )
            raise

        except requests.exceptions.HTTPError as err:
            log.error(
                "{} {}:{} over {}.".format(
                    err.response.text, self.host, self.port, self.protocol.upper()
                )
            )
            raise



import logging
from datetime import timedelta

from pypuppetdb.QueryBuilder import EqualsOperator, AndOperator
from pypuppetdb.utils import json_to_datetime

log = logging.getLogger(__name__)


class Event:
    """This object represents an event. Unless otherwise specified all
    parameters are required.

    :param node: The hostname of the node this event fired on.
    :type node: :obj:`string`
    :param status: The status for the event.
    :type status: :obj:`string`
    :param timestamp: A timestamp of when this event occured.
    :type timestamp: :obj:`string` formatted as ``%Y-%m-%dT%H:%M:%S.%fZ``
    :param hash_: The hash of the report that contains this event.
    :type hash_: :obj:`string`
    :param title: The resource title this event was fired for.
    :type title: :obj:`string`
    :param property_: The property of the resource this event was fired for.
    :type property_: :obj:`string`
    :param message: A message associated with this event.
    :type message: :obj:`string`
    :param new_value: The new value/state of the resource.
    :type new_value: :obj:`string`
    :param old_value: The old value/state of the resource.
    :type old_value: :obj:`string`
    :param type_: The type of the resource this event fired for.
    :type type_: :obj:`string`
    :param class_: The class responsible for running this event.
    :type class_: :obj:`string`
    :param execution_path: The path used to reach this particular resource.
    :type execution_path: :obj:`string`
    :param source_file: The puppet source code file containing the class.
    :type source_file: :obj:`string`
    :param line_number: The line number in the source file containing the
        definition responsible for triggering this event.
    :type line_number: :obj:`int`

    :ivar node: A :obj:`string` of this event's node certname.
    :ivar status: A :obj:`string` of this event's status.
    :ivar failed: The :obj:`bool` equivalent of `status`.
    :ivar timestamp: A :obj:`datetime.datetime` of when this event happend.
    :ivar node: The hostname of the machine this event\
        occured on.
    :ivar hash_: The hash of this event.
    :ivar item: :obj:`dict` with information about the item/resource this\
        event was triggered for.
    """

    def __init__(
        self,
        node,
        status,
        timestamp,
        hash_,
        title,
        property_,
        message,
        new_value,
        old_value,
        type_,
        class_,
        execution_path,
        source_file,
        line_number,
    ):
        self.node = node
        self.status = status
        if self.status == "failure":
            self.failed = True
        else:
            self.failed = False
        self.timestamp = json_to_datetime(timestamp)
        self.hash_ = hash_
        self.item = {
            "title": title,
            "type": type_,
            "property": property_,
            "message": message,
            "old": old_value,
            "new": new_value,
            "class": class_,
            "execution_path": execution_path,
            "source_file": source_file,
            "line_number": line_number,
        }
        self.__string = "{}[{}]/{}".format(
            self.item["type"], self.item["title"], self.hash_
        )

    def __repr__(self):
        return str(f"Event: {self.__string}")

    def __str__(self):
        return "{}".format(self.__string)

    @staticmethod
    def create_from_dict(event):
        return Event(
            node=event["certname"],
            status=event["status"],
            timestamp=event["timestamp"],
            hash_=event["report"],
            title=event["resource_title"],
            property_=event["property"],
            message=event["message"],
            new_value=event["new_value"],
            old_value=event["old_value"],
            type_=event["resource_type"],
            class_=event["containing_class"],
            execution_path=event["containment_path"],
            source_file=event["file"],
            line_number=event["line"],
        )


class Report:
    """This object represents a report. Unless otherwise specified all
    parameters are required.

    :param api: API object
    :type api: :class:`pypuppetdb.api.BaseAPI`
    :param node: The hostname of the node this report originated on.
    :type node: :obj:`string`
    :param hash_: A string uniquely identifying this report.
    :type hash_: :obj:`string`
    :param start: The start time of the agent run.
    :type start: :obj:`string` formatted as ``%Y-%m-%dT%H:%M:%S.%fZ``
    :param end: The time the agent finished its run.
    :type end: :obj:`string` formatted as ``%Y-%m-%dT%H:%M:%S.%fZ``
    :param received: The time PuppetDB received the report.
    :type received: :obj:`string` formatted as ``%Y-%m-%dT%H:%M:%S.%fZ``
    :param version: The catalog / configuration version.
    :type version: :obj:`string`
    :param format_: The catalog format version.
    :type format_: :obj:`int`
    :param agent_version: The Puppet agent version.
    :type agent_version: :obj:`string`
    :param transaction: The UUID of this transaction.
    :type transaction: :obj:`string`
    :param environment: (Optional) The environment assigned to the node that\
        submitted this report.
    :type environment: :obj:`string`
    :param status: (Optional) The status associated to this report's node.
    :type status: :obj:`string`
    :param noop: (Default `False`) A flag indicating weather the report was\
        produced by a noop run.
    :type noop: :obj:`bool`
    :param noop_pending: (Default `False`) A flag indicating weather the \
        report had pending changes produced by a noop run.
    :type noop_pending: :obj:`bool`
    :param metrics: (Optional) All metrics associated with this report.
    :type metrics: :obj:`list` containing :obj:`dict` with Metrics
    :param logs: (Optional) All logs associated with this report.
    :type logs: :obj:`list` containing :obj:`dict` of logs
    :param code_id: (Optional) Ties the catalog to the Puppet Code that\
        generated the catalog.
    :type code_id: :obj:`string`
    :param catalog_uuid: (Optional) Ties the report to the catalog used\
        from that Puppet run.
    :type catalog_uuid: :obj:`string`
    :param cached_catalog_status: (Optional) Identifies if the Puppet run\
        used a cached catalog and weather or not it was used due to an\
        error. Can be one of 'explicitly_requested', 'on_failure',\
        'not_used' not 'null'.
    :type cached_catalog_status: :obj:`string`
    :param producer: (Optional) The certname of the Puppet Server that\
        sent the report to PuppetDB
    :type producer: :obj:`string`

    :ivar node: The hostname this report originated from.
    :ivar hash_: Unique identifier of this report.
    :ivar start: :obj:`datetime.datetime` when the Puppet agent run started.
    :ivar end: :obj:`datetime.datetime` when the Puppet agent run ended.
    :ivar received: :obj:`datetime.datetime` when the report finished uploading.
    :ivar version: :obj:`string` catalog configuration version.
    :ivar format_: :obj:`int` catalog format version.
    :ivar agent_version: :obj:`string` Puppet Agent version.
    :ivar run_time: :obj:`datetime.timedelta` of **end** - **start**.
    :ivar transaction: UUID identifying this transaction.
    :ivar environment: The environment assigned to the node that submitted\
        this report.
    :ivar status: The status associated to this report's node.
    :ivar metrics: :obj:`list` containing :obj:`dict` of all metrics\
        associated with this report.
    :ivar logs: :obj:`list` containing :obj:`dict` of all logs\
        associated with this report.
    :ivar code_id: :obj:`string` used to tie a catalog to the Puppet Code\
        which generated the catalog.
    :ivar catalog_uuid: :obj:`string` used to tie this report to the catalog\
        used on this Puppet run.
    :ivar cached_catalog_status: :obj:`string` identifying if this Puppet run\
        used a cached catalog, if so weather it was a result of an error or\
        otherwise.
    :ivar producer: :obj:`string` representing the certname of the Puppet\
        Server that sent the report to PuppetDB
    """

    def __init__(
        self,
        api,
        node,
        hash_,
        start,
        end,
        received,
        version,
        format_,
        agent_version,
        transaction,
        status=None,
        metrics={},
        logs={},
        environment=None,
        noop=False,
        noop_pending=False,
        code_id=None,
        catalog_uuid=None,
        cached_catalog_status=None,
        producer=None,
    ):
        self.node = node
        self.hash_ = hash_
        self.start = json_to_datetime(start)
        self.end = json_to_datetime(end)
        self.received = json_to_datetime(received)
        self.version = version
        self.format_ = format_
        self.agent_version = agent_version
        self.run_time = self.end - self.start
        self.transaction = transaction
        self.environment = environment
        self.status = "noop" if noop and noop_pending else status
        self.metrics = metrics
        self.logs = logs
        self.code_id = code_id
        self.catalog_uuid = catalog_uuid
        self.cached_catalog_status = cached_catalog_status
        self.producer = producer
        self.__string = f"{self.hash_}"

        self.__api = api

    def __repr__(self):
        return str(f"Report: {self.__string}")

    def __str__(self):
        return "{}".format(self.__string)

    def events(self, **kwargs):
        """Get all events for this report. Additional arguments may also be
        specified that will be passed to the query function.
        """
        return self.__api.events(query=EqualsOperator("report", self.hash_), **kwargs)

    @staticmethod
    def create_from_dict(query_api, report):
        return Report(
            api=query_api,
            node=report["certname"],
            hash_=report["hash"],
            start=report["start_time"],
            end=report["end_time"],
            received=report["receive_time"],
            version=report["configuration_version"],
            format_=report["report_format"],
            agent_version=report["puppet_version"],
            transaction=report["transaction_uuid"],
            environment=report["environment"],
            status=report["status"],
            noop=report.get("noop"),
            noop_pending=report.get("noop_pending"),
            metrics=report["metrics"]["data"],
            logs=report["logs"]["data"],
            code_id=report.get("code_id"),
            catalog_uuid=report.get("catalog_uuid"),
            cached_catalog_status=report.get("cached_catalog_status"),
        )


class Fact:
    """This object represents a fact. Unless otherwise specified all
    parameters are required.

    :param node: The hostname this fact was collected from.
    :type node: :obj:`string`
    :param name: The fact's name, such as 'osfamily'
    :type name: :obj:`string`
    :param value: The fact's value, such as 'Debian'
    :type value: :obj:`string` or :obj:`int` or :obj:`dict`
    :param environment: (Optional) The fact's environment, such as\
        'production'
    :type environment: :obj:`string`

    :ivar node: :obj:`string` holding the hostname.
    :ivar name: :obj:`string` holding the fact's name.
    :ivar value: :obj:`string` or :obj:`int` or :obj:`dict` holding the\
        fact's value.
    :ivar environment: :obj:`string` holding the fact's environment
    """

    @staticmethod
    def create_from_dict(fact):
        return Fact(
            node=fact["certname"],
            name=fact["name"],
            value=fact["value"],
            environment=fact["environment"],
        )

    def __init__(self, node, name, value, environment=None):
        self.node = node
        self.name = name
        self.value = value
        self.environment = environment
        self.__string = f"{self.name}/{self.node}"

    def __repr__(self):
        return str(f"Fact: {self.__string}")

    def __str__(self):
        return "{}".format(self.__string)


class Resource:
    """This object represents a resource. Unless otherwise specified all
    parameters are required.

    :param node: The hostname this resource is located on.
    :type node: :obj:`string`
    :param name: The name of the resource in the Puppet manifest.
    :type name: :obj:`string`
    :param type_: Type of the Puppet resource.
    :type type_: :obj:`string`
    :param tags: Tags associated with this resource.
    :type tags: :obj:`list`
    :param exported: If it's an exported resource.
    :type exported: :obj:`bool`
    :param sourcefile: The Puppet manifest this resource is declared in.
    :type sourcefile: :obj:`string`
    :param sourceline: The line this resource is declared at.
    :type sourceline: :obj:`int`
    :param parameters: (Optional) The parameters this resource has been\
        declared with.
    :type parameters: :obj:`dict`
    :param environment: (Optional) The environment of the node associated\
        with this resource.
    :type environment: :obj:`string`

    :ivar node: The hostname this resources is located on.
    :ivar name: The name of the resource in the Puppet manifest.
    :ivar type_: The type of Puppet resource.
    :ivar exported: :obj:`bool` if the resource is exported.
    :ivar sourcefile: The Puppet manifest this resource is declared in.
    :ivar sourceline: The line this resource is declared at.
    :ivar parameters: :obj:`dict` with key:value pairs of parameters.
    :ivar relationships: :obj:`list` Contains all relationships to other\
        resources
    :ivar environment: :obj:`string` The environment of the node associated\
        with this resource.
    """

    def __init__(
        self,
        node,
        name,
        type_,
        tags,
        exported,
        sourcefile,
        sourceline,
        environment=None,
        parameters={},
    ):
        self.node = node
        self.name = name
        self.type_ = type_
        self.tags = tags
        self.exported = exported
        self.sourcefile = sourcefile
        self.sourceline = sourceline
        self.parameters = parameters
        self.relationships = []
        self.environment = environment
        self.__string = f"{self.type_}[{self.name}]"

    def __repr__(self):
        return "<Resource: {}>".format(self.__string)

    def __str__(self):
        return "{}".format(self.__string)

    @staticmethod
    def create_from_dict(resource):
        return Resource(
            node=resource["certname"],
            name=resource["title"],
            type_=resource["type"],
            tags=resource["tags"],
            exported=resource["exported"],
            sourcefile=resource["file"],
            sourceline=resource["line"],
            parameters=resource["parameters"],
            environment=resource["environment"],
        )


class Node:
    """This object represents a node. It additionally has some helper methods
    so that you can query for resources or facts directly from the node scope.
    Unless otherwise specified all parameters are required.

    :param api: API object.
    :type api: :class:`pypuppetdb.api.BaseAPI`
    :param name: Hostname of this node.
    :type name: :obj:`string`
    :param deactivated: (default `None`) Time this node was deactivated at.
    :type deactivated: :obj:`string` formatted as ``%Y-%m-%dT%H:%M:%S.%fZ``
    :param report_timestamp: (default `None`) Time of the last report.
    :type report_timestamp: :obj:`string` formatted as\
            ``%Y-%m-%dT%H:%M:%S.%fZ``
    :param catalog_timestamp: (default `None`) Time the last time a catalog\
            was compiled.
    :type catalog_timestamp: :obj:`string` formatted as\
            ``%Y-%m-%dT%H:%M:%S.%fZ``
    :param facts_timestamp: (default `None`) Time the last time facts were\
            collected.
    :type facts_timestamp: :obj:`string` formatted as\
            ``%Y-%m-%dT%H:%M:%S.%fZ``
    :param status_report: (default `None`) Status of latest report \
            from the node changed | unchanged | failed
    :type status: :obj:`string`
    :param noop: (Default `False`) A flag indicating whether the latest \
        report of the node was produced by a noop run.
    :type noop: :obj:`bool`
    :param noop_pending: (Default `False`) A flag indicating whether \
        the latest report of the node had pending changes \
        produced by a noop run.
    :type noop_pending: :obj:`bool`
    :param events: (default `None`) Counted events from latest Report
    :type events: :obj:`dict`
    :param unreported: (default `False`) if node is considered unreported
    :type unreported_time: :obj:`bool`
    :param unreported_time: (default `None`) Time since last report
    :type unreported_time: :obj:`string`
    :param report_environment: (default 'production') The environment of the\
            last received report for this node
    :type report_environment: :obj:`string`
    :param catalog_environment: (default 'production') The environment of the\
            last received catalog for this node
    :type catalog_environment: :obj:`string`
    :param facts_environment: (default 'production') The environment of the\
            last received fact set for this node
    :type facts_environment: :obj:`string`
    :param latest_report_hash: The hash of the latest report from this node,\
            is only available in PuppetDB 3.2 and later
    :type latest_report_hash: :obj:`string`
    :param cached_catalog_status: Cached catalog status of the last puppet run\
            on this node, possible values are 'explicitly_requested',\
            'on_failure', 'not_used' or None.
    :type cached_catalog_status: :obj:`string`

    :ivar name: Hostname of this node.
    :ivar deactivated: :obj:`datetime.datetime` when this host was\
            deactivated or `False`.
    :ivar report_timestamp: :obj:`datetime.datetime` when the last run\
            occured or `None`.
    :ivar catalog_timestamp: :obj:`datetime.datetime` last time a catalog was\
            compiled or `None`.
    :ivar facts_timestamp: :obj:`datetime.datetime` last time when facts were\
            collected or `None`.
    :ivar report_environment: :obj:`string` the environment of the last\
            received report for this node.
    :ivar catalog_environment: :obj:`string` the environment of the last\
            received catalog for this node.
    :ivar facts_environment: :obj:`string` the environment of the last\
            received fact set for this node.
    :ivar latest_report_hash: :obj:`string` the hash value of the latest\
            report the current node reported. Available in PuppetDB 3.2\
            and later.
    :ivar cached_catalog_status: :obj:`string` the status of the cached\
            catalog from the last puppet run.
    """

    def __init__(
        self,
        api,
        name,
        deactivated=None,
        expired=None,
        report_timestamp=None,
        catalog_timestamp=None,
        facts_timestamp=None,
        status_report=None,
        noop=False,
        noop_pending=False,
        events=None,
        unreported=False,
        unreported_time=None,
        report_environment="production",
        catalog_environment="production",
        facts_environment="production",
        latest_report_hash=None,
        cached_catalog_status=None,
    ):
        self.name = name
        self.events = events
        self.unreported_time = unreported_time
        self.report_timestamp = report_timestamp
        self.catalog_timestamp = catalog_timestamp
        self.facts_timestamp = facts_timestamp
        self.report_environment = report_environment
        self.catalog_environment = catalog_environment
        self.facts_environment = facts_environment
        self.latest_report_hash = latest_report_hash
        self.cached_catalog_status = cached_catalog_status

        if unreported:
            self.status = "unreported"
        elif noop and noop_pending:
            self.status = "noop"
        else:
            self.status = status_report

        if deactivated is not None:
            self.deactivated = json_to_datetime(deactivated)
        else:
            self.deactivated = False
        if expired is not None:
            self.expired = json_to_datetime(expired)
        else:
            self.expired = False
        if report_timestamp is not None:
            self.report_timestamp = json_to_datetime(report_timestamp)
        else:
            self.report_timestamp = report_timestamp
        if facts_timestamp is not None:
            self.facts_timestamp = json_to_datetime(facts_timestamp)
        else:
            self.facts_timestamp = facts_timestamp
        if catalog_timestamp is not None:
            self.catalog_timestamp = json_to_datetime(catalog_timestamp)
        else:
            self.catalog_timestamp = catalog_timestamp

        self.__api = api
        self.__string = self.name

    def __repr__(self):
        return "<Node: {}>".format(self.__string)

    def __str__(self):
        return "{}".format(self.__string)

    def facts(self, query=None, **kwargs):
        """Get all facts of this node. Additional arguments may also be
        specified that will be passed to the query function.
        """
        q = EqualsOperator("certname", self.name)
        if query:
            q = AndOperator()
            q.add(EqualsOperator("certname", self.name))
            q.add(query)

        return self.__api.facts(query=q, **kwargs)

    def fact(self, name):
        """Get a single fact from this node."""
        facts = self.facts(name=name)
        return next(fact for fact in facts)

    def resources(self, type_=None, title=None, **kwargs):
        """Get all resources of this node or all resources of the specified
        type. Additional arguments may also be specified that will be passed
        to the query function.
        """
        if type_ is None:
            resources = self.__api.resources(
                query=EqualsOperator("certname", self.name), **kwargs
            )
        elif type_ is not None and title is None:
            resources = self.__api.resources(
                type_=type_, query=EqualsOperator("certname", self.name), **kwargs
            )
        else:
            resources = self.__api.resources(
                type_=type_,
                title=title,
                query=EqualsOperator("certname", self.name),
                **kwargs,
            )
        return resources

    def resource(self, type_, title, **kwargs):
        """Get a resource matching the supplied type and title. Additional
        arguments may also be specified that will be passed to the query
        function.
        """
        resources = self.__api.resources(
            type_=type_,
            title=title,
            query=EqualsOperator("certname", self.name),
            **kwargs,
        )
        return next(resource for resource in resources)

    def reports(self, **kwargs):
        """Get all reports for this node. Additional arguments may also be
        specified that will be passed to the query function.
        """
        return self.__api.reports(query=EqualsOperator("certname", self.name), **kwargs)

    @staticmethod
    def create_from_dict(
        query_api, node, with_status, with_event_numbers, latest_events, now, unreported
    ):
        node["status_report"] = None
        node["events"] = None

        if with_status:
            if with_event_numbers:
                status = [
                    s
                    for s in latest_events
                    if s["subject"]["title"] == node["certname"]
                ]

                try:
                    node["status_report"] = node["latest_report_status"]

                    if status:
                        node["events"] = status[0]
                except KeyError:
                    if status:
                        node["events"] = status = status[0]
                        if status["successes"] > 0:
                            node["status_report"] = "changed"
                        if status["noops"] > 0:
                            node["status_report"] = "noop"
                        if status["failures"] > 0:
                            node["status_report"] = "failed"
                    else:
                        node["status_report"] = "unchanged"
            else:
                node["status_report"] = node["latest_report_status"]
                node["events"] = {
                    "successes": 0,
                    "failures": 0,
                    "noops": 0,
                }
                if node["status_report"] == "changed":
                    node["events"]["successes"] = "some"
                elif node["status_report"] == "noop":
                    node["events"]["noops"] = "some"
                elif node["status_report"] == "failed":
                    node["events"]["failures"] = "some"

            # node report age
            if node["report_timestamp"] is not None:
                try:
                    last_report = json_to_datetime(node["report_timestamp"])
                    last_report = last_report.replace(tzinfo=None)
                    unreported_border = now - timedelta(hours=unreported)
                    if last_report < unreported_border:
                        delta = now - last_report
                        node["unreported"] = True
                        node["unreported_time"] = "{}d {}h {}m".format(
                            delta.days,
                            int(delta.seconds / 3600),
                            int((delta.seconds % 3600) / 60),
                        )
                except AttributeError:
                    node["unreported"] = True

            if not node["report_timestamp"]:
                node["unreported"] = True

        return Node(
            query_api,
            name=node["certname"],
            deactivated=node["deactivated"],
            expired=node["expired"],
            report_timestamp=node["report_timestamp"],
            catalog_timestamp=node["catalog_timestamp"],
            facts_timestamp=node["facts_timestamp"],
            status_report=node["status_report"],
            noop=node.get("latest_report_noop"),
            noop_pending=node.get("latest_report_noop_pending"),
            events=node["events"],
            unreported=node.get("unreported"),
            unreported_time=node.get("unreported_time"),
            report_environment=node["report_environment"],
            catalog_environment=node["catalog_environment"],
            facts_environment=node["facts_environment"],
            latest_report_hash=node.get("latest_report_hash"),
            cached_catalog_status=node.get("cached_catalog_status"),
        )


class Catalog:
    """
    This object represents a compiled catalog from puppet. It contains\
    Resource and Edge object that represent the dependency graph. Unless\
    otherwise specified all parameters are required.

    :param node: Name of the host
    :type node: :obj:`string`
    :param edges: Edges returned from Catalog data
    :type edges: :obj:`list` containing :obj:`dict` of\
        :class:`pypuppetdb.types.Edge`
    :param resources: :class:`pypuppetdb.types.Resource` managed as of this\
        Catalog.
    :type resources: :obj:`dict` of :class:`pypuppetdb.types.Resource`
    :param version: Catalog version from Puppet (unique for each node)
    :type version: :obj:`string`
    :param transaction_uuid: A string used to match the catalog with the\
        corresponding report that was issued during the same puppet run
    :type transaction_uuid: :obj:`string`
    :param environment: The environment associated with the catalog's\
        certname.
    :type environment: :obj:`string`
    :param code_id: The string used to tie this catalog to the Puppet code\
        which generated the catalog.
    :type code_id: :obj:`string`
    :param catalog_uuid: Universally unique identifier of this catalog.
    :type catalog_uuid: :obj:`string`
    :param producer: The certname of the Puppet Server that sent the catalog\
        to PuppetDB
    :type producer: :obj:`string`

    :ivar node: :obj:`string` Name of the host
    :ivar version: :obj:`string` Catalog version from Puppet
        (unique for each node)
    :ivar transaction_uuid: :obj:`string` used to match the catalog with
        corresponding report
    :ivar edges: :obj:`list` of :obj:`Edge` The source Resource object\
        of the relationship
    :ivar environment: :obj:`string` Environment associated with the
        catalog's certname
    :ivar code_id: :obj:`string` ties the catalog to the Puppet code that\
        generated the catalog
    :ivar catalog_uuid: :obj:`string` uniquely identifying this catalog.
    :ivar producer: :obj:`string` of the Puppet Server that sent the catalog\
        to PuppetDB
    """

    def __init__(
        self,
        node,
        edges,
        resources,
        version,
        transaction_uuid,
        environment=None,
        code_id=None,
        catalog_uuid=None,
        producer=None,
    ):
        self.node = node
        self.version = version
        self.transaction_uuid = transaction_uuid
        self.environment = environment
        self.code_id = code_id
        self.catalog_uuid = catalog_uuid
        self.producer = producer

        self.resources = dict()
        for resource in resources:
            if "file" not in resource:
                resource["file"] = None
            if "line" not in resource:
                resource["line"] = None
            identifier = resource["type"] + "[" + resource["title"] + "]"
            res = Resource(
                node=node,
                name=resource["title"],
                type_=resource["type"],
                tags=resource["tags"],
                exported=resource["exported"],
                sourcefile=resource["file"],
                sourceline=resource["line"],
                parameters=resource["parameters"],
                environment=self.environment,
            )
            self.resources[identifier] = res

        self.edges = []
        for edge in edges:
            identifier_source = edge["source_type"] + "[" + edge["source_title"] + "]"
            identifier_target = edge["target_type"] + "[" + edge["target_title"] + "]"
            e = Edge(
                source=self.resources[identifier_source],
                target=self.resources[identifier_target],
                relationship=edge["relationship"],
                node=self.node,
            )
            self.edges.append(e)
            self.resources[identifier_source].relationships.append(e)
            self.resources[identifier_target].relationships.append(e)

        self.__string = f"{self.node}/{self.transaction_uuid}"

    def __repr__(self):
        return "<Catalog: {}>".format(self.__string)

    def __str__(self):
        return "{}".format(self.__string)

    def get_resources(self):
        return self.resources.values()

    def get_resource(self, resource_type, resource_title):
        identifier = resource_type + "[" + resource_title + "]"
        return self.resources[identifier]

    def get_edges(self):
        return iter(self.edges)

    @staticmethod
    def create_from_dict(catalog):
        return Catalog(
            node=catalog["certname"],
            edges=catalog["edges"]["data"],
            resources=catalog["resources"]["data"],
            version=catalog["version"],
            transaction_uuid=catalog["transaction_uuid"],
            environment=catalog["environment"],
            code_id=catalog.get("code_id"),
            catalog_uuid=catalog.get("catalog_uuid"),
        )


class Edge:
    """
    This object represents the connection between two Resource objects.
    Unless otherwise specified all parameters are required.

    :param source: The source Resource object of the relationship
    :type source: :class:`pypuppetdb.Resource`
    :param target: The target Resource object of the relationship
    :type target: :class:`pypuppetdb.Resource`
    :param relationship: Name of the Puppet Resource Relationship
    :type relationship: :obj:`string`
    :param node: The certname of the node that owns this Relationship
    :type node: :obj:`string`

    :ivar source: :obj:`Resource` The source Resource object
    :ivar target: :obj:`Resource` The target Resource object
    :ivar relationship: :obj:`string` Name of the Puppet Resource relationship
    :ivar node: :obj:`string` The name of the node that owns this relationship
    """

    def __init__(self, source, target, relationship, node=None):
        self.source = source
        self.target = target
        self.relationship = relationship
        self.node = node
        self.__string = "{} - {} - {}".format(
            self.source, self.relationship, self.target
        )

    def __repr__(self):
        return "<Edge: {}>".format(self.__string)

    def __str__(self):
        return "{}".format(self.__string)

    @staticmethod
    def create_from_dict(edge):
        identifier_source = edge["source_type"] + "[" + edge["source_title"] + "]"
        identifier_target = edge["target_type"] + "[" + edge["target_title"] + "]"
        return Edge(
            source=identifier_source,
            target=identifier_target,
            relationship=edge["relationship"],
            node=edge["certname"],
        )


class Inventory:
    """This object represents a Node Inventory entry returned from
    the Inventory endpoint.

    :param node: The certname of the node associated with the inventory.
    :type node: :obj:`string`
    :param time: The time at which PuppetDB received the facts in the
        inventory.
    :type time: :obj:`string` formatted as ``%Y-%m-%dT%H:%M:%S.%fZ``
    :param environment: The environment associated with the inventory's
        certname.
    :type environment: :obj:`string`
    :param facts: The dictionary of key-value pairs for the nodes
        assosciated facts.
    :type facts: :obj:`dict`
    :param trusted: The trusted data from the node.
    :type trusted: :obj:`dict`

    :ivar node: The certname of the node associated with the inventory.
    :ivar time: The time at which PuppetDB received the facts in the
        inventory.
    :ivar environment: The environment associated with the inventory's
        certname.
    :ivar facts: The dictionary of key-value pairs for the nodes
        assosciated facts.
    :ivar trusted: The trusted data from the node.
    """

    def __init__(self, node, time, environment, facts, trusted):
        self.node = node
        self.time = json_to_datetime(time)
        self.environment = environment
        self.facts = facts
        self.trusted = trusted
        self.__string = self.node

    def __repr__(self):
        return "<Inventory: {}>".format(self.__string)

    def __str__(self):
        return "{}".format(self.__string)

    @staticmethod
    def create_from_dict(inv):
        return Inventory(
            node=inv["certname"],
            time=inv["timestamp"],
            environment=inv["environment"],
            facts=inv["facts"],
            trusted=inv["trusted"],
        )



from pypuppetdb.api.command import CommandAPI
from pypuppetdb.api.metadata import MetadataAPI
from pypuppetdb.api.metrics import MetricsAPI
from pypuppetdb.api.pql import PqlAPI
from pypuppetdb.api.query import QueryAPI
from pypuppetdb.api.status import StatusAPI


class API(
    CommandAPI,
    MetadataAPI,
    MetricsAPI,
    PqlAPI,
    QueryAPI,
    StatusAPI,
):
    pass



import datetime


# A UTC class, see:
# http://docs.python.org/2/library/datetime.html#tzinfo-objects
class UTC(datetime.tzinfo):
    """UTC"""

    def utcoffset(self, dt):
        return datetime.timedelta(0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return datetime.timedelta(0)

    def __repr__(self):
        return "<UTC>"

    def __str__(self):
        return "UTC"

    def __unicode__(self):
        return "UTC"


def json_to_datetime(date):
    """Tranforms a JSON datetime string into a timezone aware datetime
    object with a UTC tzinfo object.

    :param date: The datetime representation.
    :type date: :obj:`string`

    :returns: A timezone aware datetime object.
    :rtype: :class:`datetime.datetime`
    """
    return datetime.datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%fZ").replace(
        tzinfo=UTC()
    )



import logging
import re
from datetime import datetime

import pypuppetdb
from pypuppetdb.QueryBuilder import EqualsOperator
from pypuppetdb.api.base import BaseAPI
from pypuppetdb.errors import APIError
from pypuppetdb.types import Node, Report

log = logging.getLogger(__name__)


class PqlAPI(BaseAPI):
    """This class provides methods that interact with the `pdb/query/v4`
    PuppetDB API endpoint.
    """

    def _pql(self, pql, request_method="GET"):
        """This method prepares a PQL query to PuppetDB. Actual making
        the HTTP request is done by _make_request().

        :param pql: PQL query
        :type pql: :obj:`string`

        :param request_method: (optional) GET or POST, the default is GET

        :raises: :class:`~pypuppetdb.errors.EmptyResponseError`

        :returns: The decoded response from PuppetDB
        :rtype: :obj:`dict` or :obj:`list`
        """

        log.debug(f"_pql called with pql={pql}, request_method={request_method}")

        pql = pql.strip()
        if not pql:
            log.error("Non-empty PQL query is required!")
            raise APIError

        payload = {}

        # PQL queries are made to the same endpoint regardless of the queried entities
        url = self._url("pql")
        payload["query"] = pql

        return self._make_request(url, request_method, payload)

    # TODO: deduplicate this - see QueryAPI.nodes()
    def pql(self, pql, with_status=False, unreported=2, with_event_numbers=True):
        """Makes a PQL (Puppet Query Language) and tries to cast results
        to a rich type. If it won't work, returns plain dicts.

        :param pql: PQL query
        :type pql: :obj:`string`

        :param with_status: (optional, only for queries for nodes) include
                           the node status in the returned nodes
        :type with_status: :bool:
        :param unreported: (optional, only for queries for nodes) amount
                           of hours when a node gets marked as unreported
        :type unreported: :obj:`None` or integer
        :param with_event_numbers: (optional, only for queries for nodes)
                           include the exact number of
                           changed/unchanged/failed/noop events when
                           with_status is set to True. If set to False
                           only "some" string is provided if there are
                           resources with such status in the last report.
                           This provides performance benefits as potentially
                           slow event-counts query is omitted completely.
        :type with_event_numbers: :bool:

        :returns: A generator yielding elements of a rich type or plain dicts
        """

        type_class = self._get_type_from_query(pql)

        if type_class == Node and (
            with_status or unreported != 2 or not with_event_numbers
        ):
            log.error(
                "with_status, unreported and with_event_numbers are used only"
                " for queries for nodes!"
            )
            raise APIError

        for element in self._pql(pql=pql):
            if type_class == Node:
                # TODO: deduplicate this - see QueryAPI.nodes()

                now = datetime.utcnow()

                latest_events = None
                if with_status and with_event_numbers:
                    latest_events = self._query(
                        "event-counts",
                        query=EqualsOperator("latest_report?", True),
                        summarize_by="certname",
                    )

                yield Node.create_from_dict(
                    self,
                    element,
                    with_status,
                    with_event_numbers,
                    latest_events,
                    now,
                    unreported,
                )

            elif type_class == Report:
                yield Report.create_from_dict(self, element)
            elif type_class:
                yield type_class.create_from_dict(element)
            else:
                yield element

    @staticmethod
    def _get_type_from_query(pql):
        """Gets a rich type of the entities returned by the given
        PQL query.

        :param pql: PQL query
        :type pql: :obj:`string`

        :return: a rich type, if this library supports it
                 otherwise - None
        """

        pql = pql.strip()
        if not pql:
            log.error("Non-empty PQL query is required!")
            raise APIError

        # in PQL the beginning of the query is the type of returned entities
        # but only if the projection is empty ([]) or there is no projection
        pattern = re.compile(r"([a-z]*?)\s*(\[])?\s*{")
        match = pattern.match(pql)

        if match:
            type_name_lowercase = match.group(1)

            # class name is capitalized
            type_name = type_name_lowercase.capitalize()

            # depluralize - remove trailing "s"
            if type_name.endswith("s"):
                type_name_singular = type_name[:-1]
            else:
                type_name_singular = type_name

            log.debug(f"Type name: {type_name_singular}")
            try:
                type_class = getattr(pypuppetdb.types, type_name_singular)
                return type_class
            except AttributeError:
                log.debug(
                    f"PQL returns entities of a type {type_name_singular},"
                    f" but it is not supported by this library yet."
                )
                return None
        else:
            log.debug("No match!")
            return None



import logging
from datetime import datetime

from pypuppetdb.QueryBuilder import EqualsOperator
from pypuppetdb.api.base import BaseAPI
from pypuppetdb.types import (
    Catalog,
    Edge,
    Event,
    Fact,
    Inventory,
    Node,
    Report,
    Resource,
)

log = logging.getLogger(__name__)


class QueryAPI(BaseAPI):
    """This class provides methods that interact with the `pdb/query/v4/*`
    PuppetDB API endpoints.
    """

    def nodes(self, unreported=2, with_status=False, with_event_numbers=True, **kwargs):
        r"""Query for nodes by either name or query. If both aren't
        provided this will return a list of all nodes. This method
        also (optionally) fetches the nodes status and (optionally)
        event counts of the latest report from puppetdb.

        :param with_status: (optional) include the node status in the\
                           returned nodes
        :type with_status: :bool:
        :param unreported: (optional) amount of hours when a node gets
                           marked as unreported
        :type unreported: :obj:`None` or integer
        :param with_event_numbers: (optional) include the exact number of\
                           changed/unchanged/failed/noop events when\
                           with_status is set to True. If set to False
                           only "some" string is provided if there are
                           resources with such status in the last report.
                           This provides performance benefits as potentially
                           slow event-counts query is omitted completely.
        :type with_event_numbers: :bool:
        :param \**kwargs: The rest of the keyword arguments are passed
                           to the _query function

        :returns: A generator yieling Nodes.
        :rtype: :class:`pypuppetdb.types.Node`
        """
        nodes = self._query("nodes", **kwargs)
        now = datetime.utcnow()

        # If we happen to only get one node back it
        # won't be inside a list so iterating over it
        # goes boom. Therefor we wrap a list around it.
        if type(nodes) == dict:
            nodes = [
                nodes,
            ]

        latest_events = None
        if with_status and with_event_numbers:
            latest_events = self._query(
                "event-counts",
                query=EqualsOperator("latest_report?", True),
                summarize_by="certname",
            )

        for node in nodes:
            yield Node.create_from_dict(
                self,
                node,
                with_status,
                with_event_numbers,
                latest_events,
                now,
                unreported,
            )

    def node(self, name):
        """Gets a single node from PuppetDB.

        :param name: The name of the node search.
        :type name: :obj:`string`

        :return: An instance of Node
        :rtype: :class:`pypuppetdb.types.Node`
        """
        nodes = self.nodes(path=name)
        return next(node for node in nodes)

    def edges(self, **kwargs):
        r"""Get the known catalog edges, formed between two resources.

        :param \**kwargs: The rest of the keyword arguments are passed
                           to the _query function.

        :returns: A generating yielding Edges.
        :rtype: :class:`pypuppetdb.types.Edge`
        """
        edges = self._query("edges", **kwargs)

        for edge in edges:
            yield Edge.create_from_dict(edge)

    def environments(self, **kwargs):
        r"""Get all known environments from Puppetdb.

        :param \**kwargs: The rest of the keyword arguments are passed
                           to the _query function.

        :returns: A list of dictionaries containing the results.
        :rtype: :obj:`list` of :obj:`dict`
        """
        return self._query("environments", **kwargs)

    def facts(self, name=None, value=None, **kwargs):
        r"""Query for facts limited by either name, value and/or query.

        :param name: (Optional) Only return facts that match this name.
        :type name: :obj:`string`
        :param value: (Optional) Only return facts of `name` that\
            match this value. Use of this parameter requires the `name`\
            parameter be set.
        :type value: :obj:`string`
        :param \**kwargs: The rest of the keyword arguments are passed
            to the _query function

        :returns: A generator yielding Facts.
        :rtype: :class:`pypuppetdb.types.Fact`
        """
        if name is not None and value is not None:
            path = f"{name}/{value}"
        elif name is not None and value is None:
            path = name
        else:
            path = None

        facts = self._query("facts", path=path, **kwargs)
        for fact in facts:
            yield Fact.create_from_dict(fact)

    def factsets(self, **kwargs):
        r"""Returns a set of all facts or for a single certname.

        :param \**kwargs: The rest of the keyword arguments are passed
                           to the _query function.

        :returns: A list of dictionaries containg the results.
        :rtype: :obj:`list` of :obj:`dict`
        """
        return self._query("factsets", **kwargs)

    def fact_contents(self, **kwargs):
        r"""To complement fact_paths(), this endpoint provides the capability
        to descend into structured facts and retreive the values associated
        with fact paths.

        :param \**kwargs: The rest of the keyword arguments are passed
                           to the _query function.

        :returns: A list of dictionaries containg the results.
        :rtype: :obj:`list` of :obj:`dict`
        """
        return self._query("fact-contents", **kwargs)

    def fact_paths(self, **kwargs):
        r"""Fact Paths are intended to be a counter-part of the fact-names
        endpoint. It provides increased granularity around structured
        facts and may be used for building GUI autocompletions or other
        applications that require a basic top-level view of fact paths.

        :param \**kwargs: The rest of the keyword arguments are passed
                           to the _query function.

        :returns: A list of dictionaries containg the results.
        :rtype: :obj:`list` of :obj:`dict`
        """
        return self._query("fact-paths", **kwargs)

    def resources(self, type_=None, title=None, **kwargs):
        r"""Query for resources limited by either type and/or title or query.
        This will yield a Resources object for every returned resource.

        :param type_: (Optional) The resource type. This can be any resource
            type referenced in\
            https://puppet.com/docs/puppet/7/type.html
        :type type_: :obj:`string`
        :param title: (Optional) The name of the resource as declared as the
            'namevar' in the Puppet Manifests. This parameter requires the\
            `type_` parameter be set.
        :type title: :obj:`string`
        :param \**kwargs: The rest of the keyword arguments are passed
            to the _query function

        :returns: A generator yielding Resources
        :rtype: :class:`pypuppetdb.types.Resource`
        """

        path = None

        if type_ is not None:
            type_ = self._normalize_resource_type(type_)

            if title is not None:
                path = f"{type_}/{title}"
            elif title is None:
                path = type_

        resources = self._query("resources", path=path, **kwargs)
        for resource in resources:
            yield Resource.create_from_dict(resource)

    def catalog(self, node):
        """Get the available catalog for a given node.

        :param node: (Required) The name of the PuppetDB node.
        :type: :obj:`string`

        :returns: An instance of Catalog
        :rtype: :class:`pypuppetdb.types.Catalog`
        """
        catalogs = self.catalogs(path=node)
        return next(catalog for catalog in catalogs)

    def catalogs(self, **kwargs):
        r"""Get the catalog information from the infrastructure based on path
        and/or query results. It is strongly recommended to include query
        and/or paging parameters for this endpoint to prevent large result
        sets or PuppetDB performance bottlenecks.

        :param \**kwargs: The rest of the keyword arguments are passed
                           to the _query function.

        :returns: A generator yielding Catalogs
        :rtype: :class:`pypuppetdb.types.Catalog`
        """

        catalogs = self._query("catalogs", **kwargs)

        if type(catalogs) == dict:
            catalogs = [
                catalogs,
            ]

        for catalog in catalogs:
            yield Catalog.create_from_dict(catalog)

    def events(self, **kwargs):
        r"""A report is made up of events which can be queried either
        individually or based on their associated report hash. It is strongly
        recommended to include query and/or paging parameters for this
        endpoint to prevent large result sets or PuppetDB performance
        bottlenecks.

        :param \**kwargs: The rest of the keyword arguments are passed
                           to the _query function

        :returns: A generator yielding Events
        :rtype: :class:`pypuppetdb.types.Event`
        """
        events = self._query("events", **kwargs)
        for event in events:
            yield Event.create_from_dict(event)

    def event_counts(self, summarize_by, **kwargs):
        r"""Get event counts from puppetdb.

        :param summarize_by: (Required) The object type to be counted on.
                             Valid values are 'containing_class', 'resource'
                             and 'certname'.
        :type summarize_by: :obj:`string`
        :param count_by: (Optional) The object type that is counted when
                         building the counts of 'successes', 'failures',
                         'noops' and 'skips'. Support values are 'certname'
                         and 'resource' (default)
        :type count_by: :obj:`string`
        :param count_filter: (Optional) A JSON query that is applied to the
                             event-counts output but before the results are
                             aggregated. Supported operators are `=`, `>`,
                             `<`, `>=`, and `<=`. Supported fields are
                             `failures`, `successes`, `noops`, and `skips`.
        :type count_filter: :obj:`string`
        :param \**kwargs: The rest of the keyword arguments are passed
                           to the _query function.

        :returns: A list of dictionaries containing the results.
        :rtype: :obj:`list`
        """
        return self._query("event-counts", summarize_by=summarize_by, **kwargs)

    def aggregate_event_counts(
        self, summarize_by, query=None, count_by=None, count_filter=None
    ):
        """Get event counts from puppetdb aggregated into a single map.

        :param summarize_by: (Required) The object type to be counted on.
                             Valid values are 'containing_class', 'resource'
                             and 'certname' or any comma-separated value
                             thereof.
        :type summarize_by: :obj:`string`
        :param query: (Optional) The PuppetDB query to filter the results.
                      This query is passed to the `events` endpoint.
        :type query: :obj:`string`
        :param count_by: (Optional) The object type that is counted when
                         building the counts of 'successes', 'failures',
                         'noops' and 'skips'. Support values are 'certname'
                         and 'resource' (default)
        :type count_by: :obj:`string`
        :param count_filter: (Optional) A JSON query that is applied to the
                             event-counts output but before the results are
                             aggregated. Supported operators are `=`, `>`,
                             `<`, `>=`, and `<=`. Supported fields are
                             `failures`, `successes`, `noops`, and `skips`.
        :type count_filter: :obj:`string`

        :returns: A dictionary of name/value results.
        :rtype: :obj:`dict`
        """
        return self._query(
            "aggregate-event-counts",
            query=query,
            summarize_by=summarize_by,
            count_by=count_by,
            count_filter=count_filter,
        )

    def fact_names(self):
        """Get a list of all known facts."""
        return self._query("fact-names")

    def reports(self, **kwargs):
        r"""Get reports for our infrastructure. It is strongly recommended
        to include query and/or paging parameters for this endpoint to
        prevent large result sets and potential PuppetDB performance
        bottlenecks.

        :param \**kwargs: The rest of the keyword arguments are passed
                           to the _query function

        :returns: A generating yielding Reports
        :rtype: :class:`pypuppetdb.types.Report`
        """
        reports = self._query("reports", **kwargs)
        for report in reports:
            yield Report.create_from_dict(self, report)

    def inventory(self, **kwargs):
        r"""Get Node and Fact information with an alternative query syntax
        for structured facts instead of using the facts, fact-contents and
        factsets endpoints for many fact-related queries.

        :param \**kwargs: The rest of the keyword arguments are passed
                           to the _query function.

        :returns: A generator yielding Inventory
        :rtype: :class:`pypuppetdb.types.Inventory`
        """
        inventory = self._query("inventory", **kwargs)
        for inv in inventory:
            yield Inventory.create_from_dict(inv)



"""
pypuppetdb PuppetDB API library
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pypuppetdb is a library to work with PuppetDB's REST API. It provides a way
to query PuppetDB and a set of additional methods and objects to make working
with PuppetDB's API and the responses easier:

    >>> from pypuppetdb import connect
    >>> db = connect()
    >>> nodes = db.nodes()
    >>> print(nodes)
    <generator object 'nodes'>
    >>> for node in nodes:
    >>>   print(node)
    host1
    host2
    ...

This will return a generator object yielding Node objects for every returned
node from PuppetDB.

To query a single node the singular db.node() can be used:

   >>> node = db.node('hostname')
   >>> print(node)
   hostname

The Node objects are a bit more special in that they can query for facts and
resources themselves. Using those methods from a node object will automatically
add a query to the request scoping the request to the node.

   >>> node = db.node('hostname')
   >>> print node.fact('osfamily')
   osfamily/hostname

We can also query for facts:

   >>> facts = db.facts('osfamily')
   >>> print(facts)
   <generator object 'facts')
   >>> for fact in facts:
   >>> print(fact)
   osfamily/host1
   osfamily/host2

That queries PuppetDB for the 'osfamily' fact and will yield Fact objects,
one per node this fact is found on.

   >>> resources = db.resources('file')

Will return a generator object containing all file resources you're managing
across your infrastructure. This is probably a bad idea if you have a big
number of nodes as the response will be huge.
"""
import logging

from pypuppetdb.api import API

logging.getLogger(__name__).addHandler(logging.NullHandler())


def connect(
    host="localhost",
    port=8080,
    ssl_verify=False,
    ssl_key=None,
    ssl_cert=None,
    timeout=10,
    protocol=None,
    url_path="/",
    username=None,
    password=None,
    token=None,
):
    """Connect with PuppetDB. This will return an object allowing you
    to query the API through its methods.

    :param host: (Default: 'localhost;) Hostname or IP of PuppetDB.
    :type host: :obj:`string`

    :param port: (Default: '8080') Port on which to talk to PuppetDB.
    :type port: :obj:`int`

    :param ssl_verify: (optional) Verify PuppetDB server certificate.
    :type ssl_verify: :obj:`bool` or :obj:`string` True, False or filesystem \
            path to CA certificate.

    :param ssl_key: (optional) Path to our client secret key.
    :type ssl_key: :obj:`None` or :obj:`string` representing a filesystem\
            path.

    :param ssl_cert: (optional) Path to our client certificate.
    :type ssl_cert: :obj:`None` or :obj:`string` representing a filesystem\
            path.

    :param timeout: (Default: 10) Number of seconds to wait for a response.
    :type timeout: :obj:`int`

    :param protocol: (optional) Explicitly specify the protocol to be used
            (especially handy when using HTTPS with ssl_verify=False and
            without certs)
    :type protocol: :obj:`None` or :obj:`string`

    :param url_path: (Default: '/') The URL path where PuppetDB is served
    :type url_path: :obj:`None` or :obj:`string`

    :param username: (optional) The username to use for HTTP basic
            authentication
    :type username: :obj:`None` or :obj:`string`

    :param password: (optional) The password to use for HTTP basic
            authentication
    :type password: :obj:`None` or :obj:`string`

    :param token: (optional) The x-auth token to use for X-Authentication
    :type token: :obj:`None` or :obj:`string`
    """
    return API(
        host=host,
        port=port,
        timeout=timeout,
        ssl_verify=ssl_verify,
        ssl_key=ssl_key,
        ssl_cert=ssl_cert,
        protocol=protocol,
        url_path=url_path,
        username=username,
        password=password,
        token=token,
    )



import logging

from pypuppetdb.api.base import BaseAPI

log = logging.getLogger(__name__)


class StatusAPI(BaseAPI):
    """This class provides methods that interact with the `status/*`
    PuppetDB API endpoints.
    """

    def status(self):
        """Get PuppetDB server status.

        :returns: A dict with the PuppetDB status information
        :rtype: :obj:`dict`
        """
        return self._query("status")



import datetime
import json
import logging

from pypuppetdb.errors import APIError

log = logging.getLogger(__name__)


class BinaryOperator:
    """
    This is a parent helper class used to create PuppetDB AST queries
    for single key-value pairs for the available operators.

    It is possible to directly declare the various types of queries
    from this class. For instance the code
    BinaryOperator('=', 'certname', 'node1.example.com') generates the
    PuppetDB query '["=", "certname", "node1.example.com"]'. It is preferred
    to use the child classes as they may have restrictions specific
    to that operator.

    See
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#binary-operators
    for more information.

    :param operator: The binary query operation performed. There is
        no value checking on this field.
    :type operator: :obj:`string`
    :param field: The PuppetDB endpoint query field. See endpoint
                  documentation for valid values.
    :type field: any
    :param value: The values of the field to match, or not match.
    :type value: any
    """

    def __init__(self, operator, field, value):
        if isinstance(value, datetime.datetime):
            value = str(value)
        self.data = [operator, field, value]

    def __repr__(self):
        return f"Query: {self}"

    def __str__(self):
        return json.dumps(self.json_data())

    def json_data(self):
        return self.data


class BooleanOperator:
    """
    This is a parent helper class used to create PuppetDB AST queries
    for available boolean queries.

    It is possible to directly declare a boolean query from this class.
    For instance the code BooleanOperator("and") will create an empty
    query '["and",]'. An error will be raised if there are no queries
    added via :func:`~pypuppetdb.QueryBuilder.BooleanOperator.add`

    See
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#binary-operators
    for more information.

    :param operator: The boolean query operation to perform.
    :type operator: :obj:`string`
    """

    def __init__(self, operator):
        self.operator = operator
        self.operations = []

    def add(self, query):
        if type(query) == list:
            for i in query:
                self.add(i)
        elif type(query) == str:
            self.operations.append(json.loads(query))
        elif isinstance(query, (BinaryOperator, InOperator, BooleanOperator)):
            self.operations.append(query.json_data())
        else:
            raise APIError(
                "Can only accept fixed-string queries, arrays " + "or operator objects"
            )

    def __repr__(self):
        return f"Query: {self}"

    def __str__(self):
        return json.dumps(self.json_data())

    def json_data(self):
        if len(self.operations) == 0:
            raise APIError("At least one query operation is required")
        return [self.operator] + self.operations


class ExtractOperator:
    """
    Queries that either do not or cannot require all the key-value pairs
    from an endpoint can use the Extract Operator as described in
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#projection-operators.

    The syntax of this operator requires a function and/or a list of fields,
    an optional standard query and an optional group by clause including a
    list of fields.
    """

    def __init__(self):
        self.fields = []
        self.query = None
        self.group_by = []

    def add_field(self, field):
        if isinstance(field, list):
            for i in field:
                self.add_field(i)
        elif isinstance(field, str):
            self.fields.append(field)
        elif isinstance(field, FunctionOperator):
            self.fields.append(field.json_data())
        else:
            raise APIError(
                "ExtractOperator.add_field only supports " "lists and strings"
            )

    def add_query(self, query):
        if self.query is not None:
            raise APIError("Only one query is supported by ExtractOperator")
        elif isinstance(query, str):
            self.query = json.loads(query)
        elif isinstance(query, (BinaryOperator, SubqueryOperator, BooleanOperator)):
            self.query = query.json_data()
        else:
            raise APIError(
                "ExtractOperator.add_query only supports "
                "strings, BinaryOperator, BooleanOperator "
                "and SubqueryOperator objects"
            )

    def add_group_by(self, field):
        if isinstance(field, list):
            for i in field:
                self.add_group_by(i)
        elif isinstance(field, str):
            if len(self.group_by) == 0:
                self.group_by.append("group_by")
            self.group_by.append(field)
        elif isinstance(field, FunctionOperator):
            if len(self.group_by) == 0:
                self.group_by.append("group_by")
            self.group_by.append(field.json_data())
        else:
            raise APIError(
                "ExtractOperator.add_group_by only supports "
                "lists, strings, and FunctionOperator objects"
            )

    def __repr__(self):
        return f"Query: {self}"

    def __str__(self):
        return json.dumps(self.json_data())

    def json_data(self):
        if len(self.fields) == 0:
            raise APIError("ExtractOperator needs at least one field")

        arr = ["extract", self.fields]

        if self.query is not None:
            arr.append(self.query)
        if len(self.group_by) > 0:
            arr.append(self.group_by)

        return arr


class FunctionOperator:
    """
    Performs an aggregate function on the result of a subquery, full
    documentation is available at
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#function.
    This object can only be used in the field list or group by list of
    an ExtractOperator object.

    :param function: The name of the function to perform.
    :type function: :obj:`str`
    :param field: The name of the field to perform the function on. All
        functions with the exception of count require this value.
    :type field: :obj:`str`
    """

    def __init__(self, function, field=None, fmt=None):
        if function not in ["count", "avg", "sum", "min", "max", "to_string"]:
            raise APIError(f"Unsupport function: {function}")
        elif function != "count" and field is None:
            raise APIError("Function {} requires a field value".format(function))
        elif function == "to_string" and fmt is None:
            raise APIError("Function {0} requires an extra 'fmt' parameter")

        self.arr = ["function", function]

        if field is not None:
            self.arr.append(field)

        if function == "to_string":
            self.arr.append(fmt)

    def __repr__(self):
        return f"Query: {self}"

    def __str__(self):
        return json.dumps(self.json_data())

    def json_data(self):
        return self.arr


class SubqueryOperator:
    """
    Performs a subquery to another puppetDB object, full
    documentation is available at
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#subquery-operators
    This object must be used in combination with the InOperator according
    to documentation.

    :param endpoint: The name of the subquery object
    :type function: :obj:`str`
    """

    def __init__(self, endpoint):
        if endpoint not in [
            "catalogs",
            "edges",
            "environments",
            "events",
            "facts",
            "fact_contents",
            "fact_paths",
            "nodes",
            "reports",
            "resources",
        ]:
            raise APIError(f"Unsupported endpoint: {endpoint}")

        self.query = None
        self.arr = [f"select_{endpoint}"]

    def add_query(self, query):
        if self.query is not None:
            raise APIError("Only one query is supported by ExtractOperator")
        else:
            self.query = True
            self.arr.append(query.json_data())

    def __repr__(self):
        return f"Query: {self}"

    def __str__(self):
        return json.dumps(self.json_data())

    def json_data(self):
        return self.arr


class InOperator:
    """
    Performs boolean compare between a field a subquery result
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#subquery-operators
    This object must be used in combination with the SubqueryOperator according
    to documentation.

    :param field: The name of the subquery object
    :type function: :obj:`str`
    """

    def __init__(self, field):
        self.query = None
        self.arr = ["in", field]

    def add_query(self, query):
        if self.query is not None:
            raise APIError("Only one query is supported by ExtractOperator")
        elif isinstance(query, str):
            self.query = True
            self.arr.append(json.loads(query))
        elif isinstance(query, (ExtractOperator, FromOperator)):
            self.query = True
            self.arr.append(query.json_data())
        else:
            raise APIError(
                "InOperator.add_query only supports "
                "strings, ExtractOperator, and"
                "FromOperator objects"
            )

    def add_array(self, values):
        if self.query is not None:
            raise APIError("Only one array is supported by the InOperator")
        elif isinstance(values, list):

            def depth(a_list):
                return (isinstance(a_list, list) and len(a_list) != 0) and max(
                    map(depth, a_list)
                ) + 1

            if depth(values) == 1:
                self.query = True
                self.arr.append(["array", values])
            else:
                raise APIError(
                    "InOperator.add_array: cannot pass in "
                    "nested arrays (or empty arrays)"
                )
        else:
            raise APIError(
                "InOperator.add_array: Ill-formatted array, "
                "must be of the format: "
                "['array', [<array values>]]"
            )

    def __repr__(self):
        return f"Query: {self}"

    def __str__(self):
        return json.dumps(self.json_data())

    def json_data(self):
        return self.arr


class FromOperator:
    """
    From contextual operator that allows for queries on the root endpoint
    or subqueries into other entities:
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#from

    Ex.)
    fr = FromOperator("facts")
    fr.add_query(EqualsOperator("foo", "bar"))
    fr.add_order_by(["certname"])
    fr.add_limit(10)

    note: only supports single entity From operations
    """

    def __init__(self, endpoint):
        valid_entities = [
            "aggregate_event_counts",
            "catalogs",
            "edges",
            "environments",
            "event_counts",
            "events",
            "facts",
            "fact_contents",
            "fact_names",
            "fact_paths",
            "nodes",
            "producers",
            "reports",
            "resources",
        ]

        if endpoint in valid_entities:
            self.endpoint = endpoint
        else:
            raise APIError(
                "Endpoint is invalid. Must be "
                "one of the following : %s" % valid_entities
            )

        self.query = None
        self.order_by = []
        self.limit = None
        self.offset = None

    def add_query(self, query):
        if self.query is not None:
            raise APIError("Only one main query is supported by FromOperator")
        elif isinstance(
            query,
            (
                InOperator,
                ExtractOperator,
                BinaryOperator,
                BooleanOperator,
                FunctionOperator,
            ),
        ):
            self.query = query.json_data()
        else:
            raise APIError("FromOperator.add_field only supports " "Operator Objects")

    def add_order_by(self, fields):
        def depth(a_list):
            return isinstance(a_list, list) and max(map(depth, a_list)) + 1

        fields_depth = depth(fields)

        if isinstance(fields, list):
            if fields_depth == 1 or fields_depth == 2:
                self.order_by = fields
            else:
                raise APIError(
                    "ExtractOperator.add_order_by only "
                    "supports lists of fields of depth "
                    "one or two: [value, <desc/asc>] or "
                    "[value]"
                )
        else:
            raise APIError(
                "ExtractOperator.add_group_by only supports "
                "lists of one or more fields"
            )

    def add_limit(self, lim):
        if isinstance(lim, int):
            self.limit = lim
        else:
            raise APIError("ExtractOperator.add_limit only supports ints")

    def add_offset(self, off):
        if isinstance(off, int):
            self.offset = off
        else:
            raise APIError("ExtractOperator.add_offset only supports ints")

    def __repr__(self):
        return f"Query: {self}"

    def __str__(self):
        return json.dumps(self.json_data())

    def json_data(self):
        if self.query is None:
            raise APIError("FromOperator needs one main query")

        arr = ["from", self.endpoint, self.query]

        if len(self.order_by) > 0:
            arr.append(["order_by", self.order_by])
        if self.limit is not None:
            arr.append(["limit", self.limit])
        if self.offset is not None:
            arr.append(["offset", self.offset])

        return arr


class EqualsOperator(BinaryOperator):
    """
    Builds an equality filter based on the supplied field-value pair as
    described
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#equality.

    In order to create the following query:

    ["=", "environment", "production"]

    The following code can be used.

    EqualsOperator('environment', 'production')

    :param field: The PuppetDB endpoint query field. See endpoint
                  documentation for valid values.
    :type field: any
    :param value: The value of the field to match, or not match.
    :type value: any
    """

    def __init__(self, field, value):
        super().__init__("=", field, value)


class GreaterOperator(BinaryOperator):
    """
    Builds a greater-than filter based on the supplied field-value pair as
    described
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#greater-than.

    In order to create the following query:

    [">", "catalog_timestamp", "2016-06-01 00:00:00"]

    The following code can be used.

    GreaterOperator('catalog_timestamp', datetime.datetime(2016, 06, 01))

    :param field: The PuppetDB endpoint query field. See endpoint
                  documentation for valid values.
    :type field: any
    :param value: Matches if the field is greater than this value.
    :type value: Number, timestamp or array
    """

    def __init__(self, field, value):
        super().__init__(">", field, value)


class LessOperator(BinaryOperator):
    """
    Builds a less-than filter based on the supplied field-value pair as
    described
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#less-than.

    In order to create the following query:

    ["<", "catalog_timestamp", "2016-06-01 00:00:00"]

    The following code can be used.

    LessOperator('catalog_timestamp', datetime.datetime(2016, 06, 01))

    :param field: The PuppetDB endpoint query field. See endpoint
                  documentation for valid values.
    :type field: any
    :param value: Matches if the field is less than this value.
    :type value: Number, timestamp or array
    """

    def __init__(self, field, value):
        super().__init__("<", field, value)


class GreaterEqualOperator(BinaryOperator):
    """
    Builds a greater-than or equal-to filter based on the supplied
    field-value pair as described
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#greater-than-or-equal-to.

    In order to create the following query:

    [">=", "facts_timestamp", "2016-06-01 00:00:00"]

    The following code can be used.

    GreaterEqualOperator('facts_timestamp', datetime.datetime(2016, 06, 01))

    :param field: The PuppetDB endpoint query field. See endpoint
                  documentation for valid values.
    :type field: any
    :param value: Matches if the field is greater than or equal to\
        this value.
    :type value: Number, timestamp or array
    """

    def __init__(self, field, value):
        super().__init__(">=", field, value)


class LessEqualOperator(BinaryOperator):
    """
    Builds a less-than or equal-to filter based on the supplied
    field-value pair as described
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#less-than-or-equal-to.

    In order to create the following query:

    ["<=", "facts_timestamp", "2016-06-01 00:00:00"]

    The following code can be used.

    LessEqualOperator('facts_timestamp', datetime.datetime(2016, 06, 01))

    :param field: The PuppetDB endpoint query field. See endpoint
                  documentation for valid values.
    :type field: any
    :param value: Matches if the field is less than or equal to\
        this value.
    :type value: Number, timestamp or array
    """

    def __init__(self, field, value):
        super().__init__("<=", field, value)


class RegexOperator(BinaryOperator):
    """
    Builds a regular expression filter based on the supplied field-value
    pair as described
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#regexp-match.

    In order to create the following query:

    ["~", "certname", "www\\d+\\.example\\.com"]

    The following code can be used.

    RegexOperator('certname', 'www\\d+\\.example\\.com')

    :param field: The PuppetDB endpoint query field. See endpoint
                  documentation for valid values.
    :type field: any
    :param value: Matches if the field matches this regular expression.
    :type value: :obj:`string`
    """

    def __init__(self, field, value):
        super().__init__("~", field, value)


class RegexArrayOperator(BinaryOperator):
    """
    Builds a regular expression array filter based on the supplied
    field-value pair. This query only works on fields with paths as
    described
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#regexp-array-match.

    In order to create the following query:

    ["~", "path", ["networking", "eth.*", "macaddress"]]

    The following code can be used.

    RegexArrayOperator('path', ["networking", "eth.*", "macaddress"])

    :param field: The PuppetDB endpoint query field. See endpoint
                  documentation for valid values.
    :type field: any
    :param value: Matches if the field matches this regular expression.
    :type value: :obj:`list`
    """

    def __init__(self, field, value):
        super().__init__("~>", field, value)


class NullOperator(BinaryOperator):
    """
    Builds a null filter based on the field and boolean value pair as
    described
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#null-is-null.
    This filter only works on field that may be null. Value may only
    be True or False.

    In order to create the following query:

    ["null?", "deactivated", true]

    The following code can be used.

    NullOperator('deactivated', True)

    :param field: The PuppetDB endpoint query field. See endpoint
                  documentation for valid values.
    :type field: any
    :param value: Matches if the field value is null (if True) or\
        not null (if False)
    :type value: :obj:`bool`
    """

    def __init__(self, field, value):
        if type(value) != bool:
            raise APIError("NullOperator value must be boolean")

        super().__init__("null?", field, value)


class AndOperator(BooleanOperator):
    """
    Builds an AND boolean filter. Only results that match ALL
    criteria from the included query strings will be returned
    from PuppetDB. Full documentation is available
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#and

    In order to create the following query:

    ["and",
      ["=", "catalog_environment", "production"],
      ["=", "facts_environment", "production"]]

    The following code can be used:

    op = AndOperator()
    op.add(EqualsOperator("catalog_environment", "production"))
    op.add(EqualsOperator("facts_environment", "production"))
    """

    def __init__(self):
        super().__init__("and")


class OrOperator(BooleanOperator):
    """
    Builds an OR boolean filter. Only results that match ANY
    criteria from the included query strings will be returned
    from PuppetDB. Full documentation is available
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#or.

    In order to create the following query:

    ["or", ["=", "name", "hostname"], ["=", "name", "architecture"]]

    The following code can be used:

    op = OrOperator()
    op.add(EqualsOperator("name", "hostname"))
    op.add(EqualsOperator("name", "architecture"))
    """

    def __init__(self):
        super().__init__("or")


class NotOperator(BooleanOperator):
    """
    Builds a NOT boolean filter. Only results that DO NOT match
    criteria from the included query strings will be returned
    from PuppetDB. Full documentation is available
    https://puppet.com/docs/puppetdb/7/api/query/v4/ast.html#not

    Unlike the other Boolean Operator objects this operator only
    accepts a single query string.

    In order to create the following query:

    ["not", ["=", "osfamily", "RedHat"]]

    The following code can be used.

    op = NotOperator()
    op.add(EqualsOperator("osfamily", "RedHat"))
    """

    def __init__(self):
        super().__init__("not")

    def add(self, query):
        if len(self.operations) > 0:
            raise APIError("This operator only accept one query string")
        elif isinstance(query, list) and len(query) > 1:
            raise APIError("This operator only accept one query string")
        super().add(query)



import logging

from pypuppetdb.api.base import BaseAPI
from pypuppetdb.errors import DoesNotComputeError

log = logging.getLogger(__name__)


class MetricsAPI(BaseAPI):
    """This class provides methods that interact with the `metrics/*`
    PuppetDB API endpoints.
    """

    def metric(self, metric=None, version=None):
        """Query for a specific metric.

        :param metric: The name of the metric we want.
        :type metric: :obj:`string`
        :param version: The version of the metric API to query. Valid values: 'v1', 'v2'
                        If not specified, then the value of self.metric_api_version
                        will be used.
        :type version: :obj:`string`

        :returns: The return of :meth:`~pypuppetdb.api.BaseAPI._query`.
        """
        version = version if version else self.metric_api_version
        if version is None or version == "v2":
            if metric is None:
                res = self._query("metrics-list")
            else:
                res = self._query("metrics", path=self._escape_metric_name(metric))

            if "error" in res:
                raise DoesNotComputeError(res["error"])
            return res["value"]
        elif version == "v1":
            return self._query("mbean", path=metric)
        else:
            raise ValueError(
                "Version specified must be 'v1' or 'v2', was given: '{}'".format(
                    version
                )
            )

    @staticmethod
    def _escape_metric_name(metric):
        """Escapes metric names so they can be used in GET requests as part of the URL.
        The new (as of v2) metrics API is backed by the Jolokia library.
        The escpaing rules for Jolokia GET requests can be found here:
        https://jolokia.org/reference/html/protocol.html#escape-rules

        :param metric: The name of the metric we want to escape.
        :type metric: :obj:`string`

        :returns: The escaped version of the metric name, safe for use in metric GET queries.
        """
        metric = metric.replace("!", r"!!")
        metric = metric.replace("/", r"!/")
        metric = metric.replace('"', r'!"')
        return metric



import logging

from pypuppetdb.api.base import BaseAPI

log = logging.getLogger(__name__)


class MetadataAPI(BaseAPI):
    """This class provides methods that interact with the `pdb/meta/*`
    PuppetDB API endpoints.
    """

    def server_time(self):
        """Get the current time of the clock on the PuppetDB server.
        :returns: An ISO-8091 formatting timestamp.
        :rtype: :obj:`string`
        """
        return self._query("server-time")[self.parameters["server_time"]]

    def current_version(self):
        """Get version information about the running PuppetDB server.

        :returns: A string representation of the PuppetDB version.
        :rtype: :obj:`string`
        """
        return self._query("version")["version"]



class APIError(Exception):
    """Our base exception the other errors inherit from."""

    pass


class ImproperlyConfiguredError(APIError):
    """This exception is thrown when the API is initialised
    and it detects incompatible configuration such as SSL turned
    on but no certificates provided."""

    pass


class EmptyResponseError(APIError):
    """Will be thrown when we did receive a response but the
    response is empty."""

    pass


class DoesNotComputeError(APIError):
    """This error will be thrown when a function is called with
    an incompatible set of optional parameters. This is the 'you are
    being a naughty developer, go read the docs' error.
    """

    pass



import hashlib
import json
import logging

import requests

from pypuppetdb.api.base import BaseAPI, COMMAND_VERSION, ERROR_STRINGS
from pypuppetdb.errors import APIError, EmptyResponseError

log = logging.getLogger(__name__)


class CommandAPI(BaseAPI):
    """This class provides methods that interact with the `pdb/cmd/*`
    PuppetDB API endpoints.
    """

    def command(self, command, payload):
        return self._cmd(command, payload)

    def _cmd(self, command, payload):
        """This method posts commands to PuppetDB. Provided a command and payload
        it will fire a request at PuppetDB. If PuppetDB can be reached and
        answers within the timeout we'll decode the response and give it back
        or raise for the HTTP Status Code PuppetDB gave back.

        :param command: The PuppetDB Command we want to execute.
        :type command: :obj:`string`

        :param command: The payload, in wire format, specific to the command.
        :type command: :obj:`dict`

        :raises: :class:`~pypuppetdb.errors.EmptyResponseError`

        :returns: The decoded response from PuppetDB
        :rtype: :obj:`dict` or :obj:`list`
        """
        log.debug("_cmd called with command: {}, data: {}".format(command, payload))

        url = self._url("cmd")

        if command not in COMMAND_VERSION:
            log.error(
                "Only {} supported, {} unsupported".format(
                    list(COMMAND_VERSION.keys()), command
                )
            )
            raise APIError

        params = {
            "command": command,
            "version": COMMAND_VERSION[command],
            "certname": payload["certname"],
            "checksum": hashlib.sha1(str(payload).encode("utf-8")).hexdigest(),  # nosec
        }

        try:
            r = self.session.post(
                url,
                params=params,
                data=json.dumps(payload, default=str),
                verify=self.ssl_verify,
                cert=(self.ssl_cert, self.ssl_key),
                timeout=self.timeout,
            )

            r.raise_for_status()

            json_body = r.json()
            if json_body is not None:
                return json_body
            else:
                del json_body
                raise EmptyResponseError

        except requests.exceptions.Timeout:
            log.error(
                "{} {}:{} over {}.".format(
                    ERROR_STRINGS["timeout"],
                    self.host,
                    self.port,
                    self.protocol.upper(),
                )
            )
            raise
        except requests.exceptions.ConnectionError:
            log.error(
                "{} {}:{} over {}.".format(
                    ERROR_STRINGS["refused"],
                    self.host,
                    self.port,
                    self.protocol.upper(),
                )
            )
            raise
        except requests.exceptions.HTTPError as err:
            log.error(
                "{} {}:{} over {}.".format(
                    err.response.text, self.host, self.port, self.protocol.upper()
                )
            )
            raise
